import numpy as np
import sklearn.linear_model as scireg
from sklearn.preprocessing import StandardScaler
from sklearn.ensemble import RandomForestRegressor
import asyncio
import threading

def euclidean_distance(point1, point2):
    return np.linalg.norm(point1 - point2)

class Calibrator:

    PRECISION_LIMIT = 50
    PRECISION_STEP = 10
    ACCEPTANCE_RADIUS = 500

    def __init__(self,CALIBRATION_RADIUS=1000):
        self.X = []
        self.Y_y = []
        self.Y_x = []
        self.reg = None
        self.reg_x = scireg.Ridge(alpha=0.5)
        self.reg_y = scireg.Ridge(alpha=0.5)
        self.current_algorithm = "Ridge"
        self.fitted = False
        self.cv_not_set = True

        self.matrix = CalibrationMatrix()
        
        self.precision_limit = self.PRECISION_LIMIT
        self.precision_step = self.PRECISION_STEP
        self.acceptance_radius = int(CALIBRATION_RADIUS/2)
        self.calibration_radius = int(CALIBRATION_RADIUS)

        self.lock = threading.Lock()
        self.calcualtion_coroutine = threading.Thread(target=self.__async_post_fit)
        self.fit_coroutines = [] 

    def __launch_fit(self):
        coroutine = threading.Thread(target=self.__async_fit)
        self.fit_coroutines.append(coroutine)
        coroutine.start()
        self.__join_finished()

    def __join_finished(self):
        for coroutine in self.fit_coroutines:
            if not coroutine.is_alive():
                coroutine.join()


    def add(self,x,y):
        self.X.append(x.flatten())
        self.Y_y.append(y[1])
        self.Y_x.append(y[0])

        self.__launch_fit()

    # This coroutine helps to asynchronously recalculate results
    def __async_fit(self):
        try:
            with self.lock:
                __tmp_X   = np.array(self.X)
                __tmp_Y_y = np.array(self.Y_y)
                __tmp_Y_x = np.array(self.Y_x)
                self.reg_x.fit(__tmp_X,__tmp_Y_x)
                self.reg_y.fit(__tmp_X,__tmp_Y_y)
                self.fitted = True
        except Exception as e:
            print(f"Exception as {e}")

    # This coroutine helps to asynchronously recalculate results
    def __async_post_fit(self):
        try:
            tmp_fixations_x = scireg.LassoCV(cv=50,max_iter=5000)
            tmp_fixations_y = scireg.LassoCV(cv=50,max_iter=5000)

            __tmp_X   = np.array(self.X)
            __tmp_Y_y = np.array(self.Y_y)
            __tmp_Y_x = np.array(self.Y_x)

            tmp_fixations_x.fit(__tmp_X,__tmp_Y_x)
            tmp_fixations_y.fit(__tmp_X,__tmp_Y_y)

            with self.lock:
                self.fixations_x = tmp_fixations_x
                self.fixations_y = tmp_fixations_y
                self.fitted = True
                self.current_algorithm = "LassoCV"
        except Exception as e:
            print(f"Exception as {e}")
            self.cv_not_set = True

    def post_fit(self):
        if self.cv_not_set:
            self.calcualtion_coroutine.start()
            self.cv_not_set = False

    def whichAlgorithm(self):
        with self.lock:
            return self.current_algorithm

    def predict(self,x):
        with self.lock:
            if self.fitted:
                x = x.flatten()
                x = x.reshape(1, -1)
                y_x = self.reg_x.predict(x)[0]
                y_y = self.reg_y.predict(x)[0]
                return np.array([y_x,y_y])
            else:
                return np.array([0.0,0.0])

    def movePoint(self):
        self.matrix.movePoint()

    def getCurrentPoint(self,width,heigth):
        return self.matrix.getCurrentPoint(width,heigth)

    def updMatrix(self,points):
        return self.matrix.updMatrix(points)

    def unfit(self):
        self.acceptance_radius = self.ACCEPTANCE_RADIUS
        self.calibration_radius = self.CALIBRATION_RADIUS
        self.fitted = False

    def increase_precision(self):
        if self.acceptance_radius > self.precision_limit:
            self.acceptance_radius -= self.precision_step
        if self.calibration_radius > self.precision_limit and self.acceptance_radius < self.calibration_radius:
            self.calibration_radius -= self.precision_step

    def insideClbRadius(self,point,width,height):
        return euclidean_distance(point,self.getCurrentPoint(width,height)) < self.calibration_radius
    
    def insideAcptcRadius(self,point,width,height):
        return euclidean_distance(point,self.getCurrentPoint(width,height)) < self.acceptance_radius
    
class CalibrationMatrix:

    def __init__(self):

        self.iterator = 0
        self.points = np.array([[1,0.5],[0.75,0.5],[0.5,0.5],[0.25,0.5],[0.0,0.5],
                                [1.0,1.0],[0.75,1.0],[0.5,1.0],[0.25,1.0],[0.0,1.0],
                                [1.0,0.0],[0.75,0.0],[0.5,0.0],[0.25,0.0],[0.0,0.0],
                                [0.0,0.75],[0.75,0.75],[0.5,0.75],[0.25,0.75],[0.0,0.75],
                                [1.0,0.25],[0.75,0.25],[0.5,0.25],[0.25,0.25],[0.0,0.25]])
        pass

    def updMatrix(self,points):
        self.points = points
        self.iterator = 0

    def movePoint(self):
        self.iterator += 1
        self.iterator %= len(self.points)

    def getCurrentPoint(self,width=1.0,height=1.0):
        it = self.iterator
        return np.array([self.points[it,0] * width, self.points[it,1] * height])


import math
import numpy as np

from scipy import signal

class Center():
    """Helper representing Center of object with x,y,width,height"""

    def __init__(self,x,y,width,height):
        self.x = (x + width)/2
        self.y = (y + height)/2

class Screen():
    """Helper representing screen by making box of width,height"""

    def __init__(self,width,height):
        self.x = 0
        self.y = 0
        self.width = width
        self.height= height

    def getCenter(self):
        """Function returnig center of the screen"""
        return Center(self.x,self.y,self.width, self.height)

class ScreenROI():
    """Helper representing ROI by making box of width,height with x and y offset"""

    def __init__(self,x,y,width,height):
        self.x = x 
        self.y = y
        self.width = width
        self.height= height

    def setCenter(self,x, y):
        """Function allowing to move ROI by changing its center"""
        self.x = x - self.width/2
        self.y = y - self.height/2

    def getCenter(self):
        """Function returnig center of the ROI"""
        return Center(self.x, self.y, self.width, self.height)

    def getBoundaries(self):
        """Function returnig boundaries of the ROI"""
        return (self.x,self.y,self.width, self.height)


class Display():
    """Helper representing Display by making box of width,height with x and y offset"""

    def __init__(self,width,height,offset_x,offset_y):
        self.width = width
        self.height = height
        self.offset_x = offset_x
        self.offset_y = offset_y

        # self.buffor = Buffor(20)

    def getCenter(self):
        """Function returnig center of the Display"""
        return self.__center



import eyeGestures.screenTracker.dataPoints as dp
import eyeGestures.screenTracker.screenTracker as scrtr

# virtual tracking ScreenProcessor
SCREEN_WIDTH = 500
SCREEN_HEIGHT = 500

# region of interest inside the virtual ScreenProcessor
ROI_WIDTH = 50
ROI_HEIGHT = 50

# size of real display
DISPLAY_WIDTH = 1200
DISPLAY_HEIGHT = 1800

####################### ScreenProcessor 2 display Tests #################################


def test_screen2display_1_1():
    """[TEST]"""
    point = (1, 1)

    roi = dp.ScreenROI(0, 0, ROI_WIDTH, ROI_HEIGHT)
    display = dp.Display(DISPLAY_WIDTH, DISPLAY_HEIGHT, 0, 0)

    tracker = scrtr.ScreenProcessor()
    p = tracker.screen2display(point, roi, display)

    assert p == (point[0]/ROI_WIDTH * DISPLAY_WIDTH,
                 point[1]/ROI_HEIGHT * DISPLAY_HEIGHT)


def test_screen2display_55_1():
    """[TEST]"""
    point = (55, 1)

    roi = dp.ScreenROI(0, 0, ROI_WIDTH, ROI_HEIGHT)
    display = dp.Display(DISPLAY_WIDTH, DISPLAY_HEIGHT, 0, 0)

    tracker = scrtr.ScreenProcessor()
    p = tracker.screen2display(point, roi, display)

    assert p == (DISPLAY_WIDTH,
                 point[1]/ROI_HEIGHT * DISPLAY_HEIGHT)


def test_screen2display_55_55():
    """[TEST]"""
    point = (55, 55)

    roi = dp.ScreenROI(0, 0, ROI_WIDTH, ROI_HEIGHT)
    display = dp.Display(DISPLAY_WIDTH, DISPLAY_HEIGHT, 0, 0)

    tracker = scrtr.ScreenProcessor()
    p = tracker.screen2display(point, roi, display)

    assert p == (DISPLAY_WIDTH,
                 DISPLAY_HEIGHT)


def test_screen2display_0_0():
    """[TEST]"""
    point = (0, 0)

    roi = dp.ScreenROI(0, 0, ROI_WIDTH, ROI_HEIGHT)
    display = dp.Display(DISPLAY_WIDTH, DISPLAY_HEIGHT, 0, 0)

    tracker = scrtr.ScreenProcessor()
    p = tracker.screen2display(point, roi, display)

    assert p == (0, 0)


def test_screen2display_1_1_pos_50_50():
    """[TEST]"""
    point = (51, 51)
    pos = (50, 50)

    roi = dp.ScreenROI(pos[0], pos[1], ROI_WIDTH, ROI_HEIGHT)
    display = dp.Display(DISPLAY_WIDTH, DISPLAY_HEIGHT, 0, 0)

    tracker = scrtr.ScreenProcessor()
    p = tracker.screen2display(point, roi, display)

    assert p == ((point[0] - pos[0])/ROI_WIDTH * DISPLAY_WIDTH,
                 (point[1] - pos[1])/ROI_HEIGHT * DISPLAY_HEIGHT)


def test_screen2display_55_55_pos_50_50():
    """[TEST]"""
    point = (105, 105)
    pos = (50, 50)

    roi = dp.ScreenROI(pos[0], pos[1], ROI_WIDTH, ROI_HEIGHT)
    display = dp.Display(DISPLAY_WIDTH, DISPLAY_HEIGHT, 0, 0)

    tracker = scrtr.ScreenProcessor()
    p = tracker.screen2display(point, roi, display)

    assert p == (DISPLAY_WIDTH,
                 DISPLAY_HEIGHT)


def test_screen2display_0_0_pos_50_50():
    """[TEST]"""
    point = (0, 0)
    pos = (50, 50)

    roi = dp.ScreenROI(pos[0], pos[1], ROI_WIDTH, ROI_HEIGHT)
    display = dp.Display(DISPLAY_WIDTH, DISPLAY_HEIGHT, 0, 0)

    tracker = scrtr.ScreenProcessor()
    p = tracker.screen2display(point, roi, display)

    assert p == (0, 0)



"""Module providing a Gaze Events."""


class Gevent:
    """Class representing gaze event, with tracked points scaled to screen, blink and fixation."""

    def __init__(self,
                 point,
                 blink,
                 fixation,
                 l_eye = None,
                 r_eye = None,
                 screen_man = None,
                 roi = None,
                 edges = None,
                 cluster = None,
                 context = None,
                 saccades = False):

        self.point = point
        self.blink = blink
        self.fixation = fixation
        self.saccades = saccades

        # ALL DEBUG DATA
        self.roi = roi
        self.edges = edges
        self.l_eye = l_eye
        self.r_eye = r_eye
        self.cluster = cluster
        self.context = context
        self.screen_man = screen_man


class Cevent:
    """Class representing gaze event, with tracked points scaled to screen, blink and fixation."""

    def __init__(self,
                 point,
                 acceptance_radius,
                 calibration_radius,
                 calibration = False):

        self.point = point
        self.acceptance_radius = acceptance_radius
        self.calibration_radius = calibration_radius
        self.calibration = calibration



import time


class CalibrationPositions:
    LEFT = "left"
    RIGHT = "right"
    TOP = "top"
    BOTTOM = "bottom"

class Calibrator:

    def __init__(self, width, height, start_x, start_y):
        self.width = width
        self.height = height
        self.start_x = start_x
        self.start_y = start_y

        self.prev_x = 0
        self.prev_y = 0

        self.calibration_margin = 200
        self.calibration_steps = []
        self.__set_order()

        self.calibrate_left = False
        self.calibrate_right = False
        self.calibrate_top = False
        self.calibrate_bottom = False

        self.calibration = False
        self.drawn = False
        self.prev_point = None
        self.last_calib = time.time()
        pass

    def __add_left(self):
        self.calibration_steps.append(CalibrationPositions.LEFT)
        return self
    def __add_right(self):
        self.calibration_steps.append(CalibrationPositions.RIGHT)
        return self

    def __add_top(self):
        self.calibration_steps.append(CalibrationPositions.TOP)
        return self

    def __add_bottom(self):
        self.calibration_steps.append(CalibrationPositions.BOTTOM)
        return self

    def __set_order(self):
        if self.start_x < self.width/2 and CalibrationPositions.LEFT not in self.calibration_steps:
            self.__add_left().__add_right()
        elif self.start_x > self.width/2 and CalibrationPositions.RIGHT not in self.calibration_steps:
            self.__add_right().__add_left()

        if self.start_y < self.height/2 and CalibrationPositions.TOP not in self.calibration_steps:
            self.__add_top().__add_bottom()
        elif self.start_y > self.height/2 and CalibrationPositions.BOTTOM not in self.calibration_steps:
            self.__add_bottom().__add_top()

    def add_recalibrate(self,recalibrate_step):

        if recalibrate_step not in self.calibration_steps:
            self.calibration_steps.append(recalibrate_step)

    def get_current_point(self):
        if len(self.calibration_steps) > 0:
            if CalibrationPositions.LEFT == self.calibration_steps[0]:
                return (self.calibration_margin, int(self.height/2))
            elif CalibrationPositions.RIGHT == self.calibration_steps[0]:
                return (self.width - self.calibration_margin, int(self.height/2))
            elif CalibrationPositions.TOP == self.calibration_steps[0]:
                return (int(self.width/2), self.calibration_margin)
            elif CalibrationPositions.BOTTOM == self.calibration_steps[0]:
                return (int(self.width/2), self.height - self.calibration_margin)
        else:
            return (0,0)

    def calibrate(self,x,y,fix):

        if abs(x - self.width/2) > 150 and self.prev_point in [CalibrationPositions.TOP, CalibrationPositions.BOTTOM]:
            if x - self.width/2 < 0:
                self.add_recalibrate(CalibrationPositions.LEFT)
            else:
                self.add_recalibrate(CalibrationPositions.RIGHT)

        if abs(y - self.height/2) > 150  and self.prev_point in [CalibrationPositions.LEFT, CalibrationPositions.RIGHT]:
            if y - self.height/2 < 0:
                self.add_recalibrate(CalibrationPositions.TOP)
            else:
                self.add_recalibrate(CalibrationPositions.BOTTOM)

        self.prev_y = y
        self.prev_x = x
        if len(self.calibration_steps) <= 0:
            self.prev_point = None
            return False

        fixation_thresh = 0.3
        if fix > fixation_thresh and (time.time() - self.last_calib) > 5.0:
            if CalibrationPositions.LEFT == self.calibration_steps[0] and x < self.calibration_margin:
                if CalibrationPositions.LEFT in self.calibration_steps:
                    self.calibration_steps.remove(CalibrationPositions.LEFT)
                self.prev_point = CalibrationPositions.LEFT
                self.drawn = False
                self.last_calib = time.time()
                return True
            elif CalibrationPositions.RIGHT == self.calibration_steps[0] and x > self.width - self.calibration_margin:
                if CalibrationPositions.RIGHT in self.calibration_steps:
                    self.calibration_steps.remove(CalibrationPositions.RIGHT)
                self.prev_point = CalibrationPositions.RIGHT
                self.drawn = False
                self.last_calib = time.time()
                return True
            elif CalibrationPositions.TOP == self.calibration_steps[0] and y < self.calibration_margin:
                if CalibrationPositions.TOP in self.calibration_steps:
                    self.calibration_steps.remove(CalibrationPositions.TOP)
                self.prev_point = CalibrationPositions.TOP
                self.drawn = False
                self.last_calib = time.time()
                return True
            elif CalibrationPositions.BOTTOM == self.calibration_steps[0] and y > self.height - self.calibration_margin:
                if CalibrationPositions.BOTTOM in self.calibration_steps:
                    self.calibration_steps.remove(CalibrationPositions.BOTTOM)
                self.prev_point = CalibrationPositions.BOTTOM
                self.drawn = False
                self.last_calib = time.time()
                return True
            else:
                # TODO: somewhere here is bug breaking entire program
                self.last_calib = time.time()
                self.drawn = False
                self.prev_point = None

                if self.calibration_steps[0] in [CalibrationPositions.RIGHT,CalibrationPositions.LEFT]:
                    if x < self.width/2:
                        if CalibrationPositions.RIGHT in self.calibration_steps:
                            self.calibration_steps.remove(CalibrationPositions.RIGHT)
                        self.calibration_steps.insert(0,CalibrationPositions.RIGHT)
                    else:
                        if CalibrationPositions.LEFT in self.calibration_steps:
                            self.calibration_steps.remove(CalibrationPositions.LEFT)
                        self.calibration_steps.insert(0,CalibrationPositions.LEFT)
                    return True

                if self.calibration_steps[0] is CalibrationPositions.TOP:
                    self.calibration_steps.insert(0,CalibrationPositions.BOTTOM)
                    return True
                else:
                    self.calibration_steps.insert(0,CalibrationPositions.TOP)
                    return True

        self.prev_point = None
        return False

    def calibrated(self):
        return len(self.calibration_steps) <= 0





from eyeGestures.face import FaceFinder, Face
from eyeGestures.Fixation import Fixation
from eyeGestures.gazeEstimator import GazeTracker
import eyeGestures.screenTracker.dataPoints as dp
from eyeGestures.calibration_v1 import Calibrator as Calibrator_v1
from eyeGestures.calibration_v2 import Calibrator as Calibrator_v2
from eyeGestures.gevent import Gevent, Cevent
from eyeGestures.utils import timeit, Buffor
import numpy as np
import pickle
import time
import cv2

VERSION = "3.0.0"


class EyeGestures_v3:
    """Main class for EyeGesture tracker. It configures and manages entire algorithm"""

    def __init__(self, calibration_radius = 1000):
        self.calibration_radius = calibration_radius 

        self.clb = dict() # Calibrator_v2()
        self.cap = None

        self.calibration = dict()

        self.average_points = dict()
        self.iterator = dict()
        self.filled_points= dict()
        self.enable_CN = False
        self.calibrate_gestures = False

        self.finder = FaceFinder()
        self.face = Face()

        # this has to be contexted
        self.prev_timestamp     = dict()
        self.prev_point         = dict()
        self.fix                = dict()
        self.velocity_max       = dict()
        self.velocity_min       = dict()
        self.fixationTracker    = dict()

    def saveModel(self, context = "main"):
        if context in self.clb:
            return pickle.dumps(self.clb[context])

    def loadModel(self,model, context = "main"):
        self.clb[context] = pickle.loads(model)

    def uploadCalibrationMap(self,points,context = "main"):
        self.addContext(context)
        self.clb[context].updMatrix(np.array(points))

    def getLandmarks(self, frame):

        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        frame = cv2.flip(frame,1)
        # frame = cv2.resize(frame, (360, 640))

        try:
            # decoupling from V1
            self.face.process(
                frame,
                self.finder.find(frame)
            )
            l_eye = self.face.getLeftEye()
            r_eye = self.face.getRightEye()
            l_eye_landmarks = l_eye.getLandmarks()
            r_eye_landmarks = r_eye.getLandmarks()
            blink = l_eye.getBlink() and r_eye.getBlink()
            
            # eye_events = np.array([event.blink,event.fixation]).reshape(1, 2)
            key_points = np.concatenate((l_eye_landmarks,r_eye_landmarks))
            return key_points, blink
        except:
            pass
        return np.array([]), 0

    def whichAlgorithm(self,context="main"):
        if context in self.clb:
            return self.clb[context].whichAlgorithm()
        else:
            return "None"

    def reset(self, context = "main"):
        self.filled_points[context] = 0
        if context in self.clb:
           self.addContext(context)

    def setFixation(self,fix):
        self.fix = fix

    def addContext(self, context):
        if context not in self.clb:
            self.clb[context] = Calibrator_v2(self.calibration_radius)
            self.average_points[context] = Buffor(5)
            self.iterator[context] = 0
            self.average_points[context] = np.zeros((20,2))
            self.filled_points[context] = 0
            self.calibration[context] = False
            self.prev_timestamp[context] = time.time()
            self.prev_point[context] = np.array((0.0,0.0))
            self.fix[context] = 0.8
            self.velocity_max[context] = 0
            self.velocity_min[context] = 100000000
            self.fixationTracker[context] = Fixation(0,0,100)


    def step(self, frame, calibration, width, height, context="main"):
        self.addContext(context)

        self.calibration[context] = calibration

        key_points, blink = self.getLandmarks(frame)


        y_point = self.clb[context].predict(key_points)
        self.average_points[context][1:,:] = self.average_points[context][:(self.average_points[context].shape[0] - 1),:]
        self.average_points[context][0,:] = y_point

        if self.filled_points[context] < self.average_points[context].shape[0] and (y_point != np.array([0.0,0.0])).any():
            self.filled_points[context] += 1
        averaged_point = np.sum(self.average_points[context][:,:],axis=0)/(self.filled_points[context])

        fixation = self.fixationTracker[context].process(
            averaged_point[0], averaged_point[1])

        duration = time.time() - self.prev_timestamp[context]
        velocity = abs(averaged_point - self.prev_point[context])/duration
        velocity = np.sqrt(velocity[0]**2+velocity[1]**2)
        self.prev_point[context] = averaged_point
        self.prev_timestamp[context] = time.time()
 
        self.velocity_max[context] = max(self.velocity_max[context],velocity)
        self.velocity_min[context] = min(self.velocity_min[context],velocity)

        saccades = velocity > (self.velocity_max[context]+self.velocity_min[context])/4

        if self.calibration[context] and (self.clb[context].insideClbRadius(averaged_point,width,height) or self.filled_points[context] < self.average_points[context].shape[0] * 10):
            self.clb[context].add(key_points,self.clb[context].getCurrentPoint(width,height))
        else: 
            self.clb[context].post_fit()

        if self.calibration[context] and self.clb[context].insideAcptcRadius(averaged_point,width,height):
            self.iterator[context] += 1
            if self.iterator[context] > 10:
                self.iterator[context] = 0
                self.clb[context].movePoint()

        gevent = Gevent(
            point=averaged_point,
            blink=blink,
            fixation=fixation,
            saccades=saccades
        )
        cevent = Cevent(self.clb[context].getCurrentPoint(width,height),self.clb[context].acceptance_radius, self.clb[context].calibration_radius)
        return (gevent, cevent)


class EyeGestures_v2:
    """Main class for EyeGesture tracker. It configures and manages entire algorithm"""

    def __init__(self, calibration_radius = 1000):
        self.monitor_width  = 1
        self.monitor_height = 1
        self.calibration_radius = calibration_radius 

        self.clb = dict() # Calibrator_v2()
        self.cap = None
        self.gestures = EyeGestures_v1(285,115,200,100)

        self.calibration = dict()

        self.CN = 5

        self.average_points = dict()
        self.iterator = dict()
        self.filled_points= dict()
        self.enable_CN = False
        self.calibrate_gestures = False

        self.fix = 0.8

    def saveModel(self, context = "main"):
        if context in self.clb:
            return pickle.dumps(self.clb[context])

    def loadModel(self,model, context = "main"):
        self.clb[context] = pickle.loads(model)

    def uploadCalibrationMap(self,points,context = "main"):
        self.addContext(context)
        self.clb[context].updMatrix(np.array(points))

    def getLandmarks(self, frame, calibrate = False, context="main"):

        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        frame = cv2.flip(frame,1)
        # frame = cv2.resize(frame, (360, 640))

        event, cevent = self.gestures.step(
            frame,
            context,
            calibrate, # set calibration - switch to False to stop calibration
            self.monitor_width,
            self.monitor_height,
            0, 0, self.fix, 100)

        if event is not None or cevent is not None: 
            cursor_x, cursor_y = event.point[0],event.point[1]
            l_eye_landmarks = event.l_eye.getLandmarks()
            r_eye_landmarks = event.r_eye.getLandmarks()

            cursors = np.array([cursor_x,cursor_y]).reshape(1, 2)
            eye_events = np.array([event.blink,event.fixation]).reshape(1, 2)
            key_points = np.concatenate((cursors,l_eye_landmarks,r_eye_landmarks,eye_events))
            return np.array((cursor_x, cursor_y)), key_points, event.blink, event.fixation, cevent
        return np.array((0.0, 0.0)), np.array([]), 0, 0, None

    def whichAlgorithm(self,context="main"):
        if context in self.clb:
            return self.clb[context].whichAlgorithm()
        else:
            return "None"

    def setClassicImpact(self,impact):
        self.CN = impact

    def reset(self, context = "main"):
        self.filled_points[context] = 0
        if context in self.clb:
           self.addContext(context)

    def setFixation(self,fix):
        self.fix = fix

    def setClassicalImpact(self,CN):
        self.CN = CN

    def enableCNCalib(self):
        self.enable_CN = True

    def disableCNCalib(self):
        self.enable_CN = False

    def addContext(self, context):
        if context not in self.clb:
            self.clb[context] = Calibrator_v2(self.calibration_radius)
            self.average_points[context] = Buffor(20)
            self.iterator[context] = 0
            self.average_points[context] = np.zeros((5,2))
            self.filled_points[context] = 0
            self.calibration[context] = False


    def step(self, frame, calibration, width, height, context="main"):
        self.addContext(context)

        self.calibration[context] = calibration
        self.monitor_width = width
        self.monitor_height = height

        classic_point, key_points, blink, fixation, cevent = self.getLandmarks(frame,
                                                                               self.calibrate_gestures and self.enable_CN,
                                                                               context = context)

        if cevent is None:
            return (None, None)

        margin = 10
        if (classic_point[0] <= margin) and self.calibration[context]:
            self.calibrate_gestures = cevent.calibration
        elif (classic_point[0] >= width - margin) and self.calibration[context]:
            self.calibrate_gestures = cevent.calibration
        elif (cevent.point[1] <= margin) and self.calibration[context]:
            self.calibrate_gestures = cevent.calibration
        elif (classic_point[1] >= height - margin) and self.calibration[context]:
            self.calibrate_gestures = cevent.calibration
        else:
            self.calibrate_gestures = False

        y_point = self.clb[context].predict(key_points)
        self.average_points[context][1:,:] = self.average_points[context][:(self.average_points[context].shape[0] - 1),:]
        if fixation <= self.fix:
            self.average_points[context][0,:] = y_point

        if self.filled_points[context] < self.average_points[context].shape[0] and (y_point != np.array([0.0,0.0])).any():
            self.filled_points[context] += 1
        averaged_point = (np.sum(self.average_points[context][:,:],axis=0) + (classic_point * self.CN))/(self.filled_points[context] + self.CN)
        
        if self.calibration[context] and (self.clb[context].insideClbRadius(averaged_point,width,height) or self.filled_points[context] < self.average_points[context].shape[0] * 10):
            self.clb[context].add(key_points,self.clb[context].getCurrentPoint(width,height))
        else: 
            self.clb[context].post_fit()

        if self.calibration[context] and self.clb[context].insideAcptcRadius(averaged_point,width,height):
            self.iterator[context] += 1
            if self.iterator[context] > 10:
                self.iterator[context] = 0
                self.clb[context].movePoint()

        gevent = Gevent(averaged_point,blink,fixation)
        cevent = Cevent(self.clb[context].getCurrentPoint(width,height),self.clb[context].acceptance_radius, self.clb[context].calibration_radius)
        return (gevent, cevent)

class EyeGestures_v1:
    """Main class for EyeGesture tracker. It configures and manages entier algorithm"""

    def __init__(self,
                 roi_x=225,
                 roi_y=105,
                 roi_width=80,
                 roi_height=15):

        screen_width = 500
        screen_height = 500
        height = 250
        width = 250

        roi_x = roi_x % screen_width
        roi_y = roi_y % screen_height
        roi_width = roi_width % screen_width
        roi_height = roi_height % screen_height

        self.screen_width = screen_width
        self.screen_height = screen_height

        self.screen = dp.Screen(
            screen_width,
            screen_height)

        self.gaze = GazeTracker(screen_width,
                                screen_height,
                                width,
                                height,
                                roi_x,
                                roi_y,
                                roi_width,
                                roi_height)

        self.calibrators = dict()
        self.calibrate = False

    def getFeatures(self, image):
        """[NOT RECOMMENDED] Function allowing for extraction of gaze features from image"""

        return self.gaze.getFeatures(image)

    # @timeit
    # 0.011 - 0.015 s for execution
    def step(self, image,
                 context,
                 calibration,
                 display_width,
                 display_height,
                 display_offset_x=0,
                 display_offset_y=0,
                 fixation_freeze=0.7,
                 freeze_radius=20,
                 offset_x=0,
                 offset_y=0):
        """Function performing estimation and returning event object"""

        display = dp.Display(
            display_width,
            display_height,
            display_offset_x,
            display_offset_y
        )

        # allow for random recalibration shot
        event = self.gaze.estimate(image,
                                  display,
                                  context,
                                  calibration,
                                  fixation_freeze,
                                  freeze_radius,
                                  offset_x,
                                  offset_y)
        cevent = None

        if event is not None:
            cursor_x,cursor_y = event.point[0],event.point[1]
            if context in self.calibrators:
                self.calibrate = self.calibrators[context].calibrate(cursor_x,cursor_y,event.fixation)
            else:
                self.calibrators[context] = Calibrator_v1(display_width,display_height,cursor_x,cursor_y)
                self.calibrate = self.calibrators[context].calibrate(cursor_x,cursor_y,event.fixation)

            cpoint = self.calibrators[context].get_current_point()
            calibration = self.calibrate and cpoint != (0,0)
            cevent = Cevent(cpoint,100, 100, calibration)

        return (event, cevent)


from sklearn.cluster import DBSCAN
import numpy as np

class Cluster:
    """Class representing one ROI cluster"""

    def __init__(self, label, points):
        
        self.label    = label
        self.points   = np.array(points)
        self.weight   = len(self.points)
        self.__centroid = self.centroid(points)

        x,y,w,h = self.boundaries(points)

        self.x =  x
        self.y =  y
        self.w =  w
        self.h =  h

    def centroid(self,points):
        """Function calculating and returning center of ROI"""
        
        x,y,w,h = self.boundaries(points)

        c_x, c_y = (x+w/2,y+h/2)

        x,y = sum(points)/len(points)

        c_x, c_y = (c_x + x)/2, (c_y + y)/2
        return (c_x, c_y)

    def boundaries(self,points):
        """Function calculating and returning boundaries of cluster"""

        x = np.min(points[:,0]) 
        width = abs(np.max(points[:,0]) - x) 

        y = np.min(points[:,1]) 
        height = abs(np.max(points[:,1]) - y) 

        return (x,y,width,height)
    
    def getBoundaries(self):
        """Function returning boundaries of cluster"""
        return (self.x,self.y,self.w,self.h)
    
    def getCenter(self):
        """Function returning center of ROI"""
        return (self.__centroid)
    
class Clusters:
    """Algorithm dividing tracked points into clusters so it can be estimated were is center of ROI"""

    def __init__(self,buffer):
        buffer = np.array(buffer)
        
        self.head = len(buffer) - 1
        self.main_cluster = None
        self.clusters = []

        dbScan     = DBSCAN(eps=12, min_samples=3)
        clustering = dbScan.fit(buffer)

        labels = clustering.labels_
        unique_labels = set(labels) - {-1}

        for u_label in unique_labels:
            core_sample_indices = clustering.core_sample_indices_
            cluster_points = buffer[core_sample_indices][labels[clustering.core_sample_indices_] == u_label]
            self.clusters.append(Cluster(u_label,cluster_points))

        # return sorted(self.clusters, key=lambda cluster: cluster.weight)[-1]
        if len(self.clusters) > 0:
            self.main_cluster = self.clusters[
                    labels[self.head]
                ]
        else:
            self.main_cluster = None

    def clearPoints(self):
        """[Possibly dead code] Function reseting cluster"""
        self.head = 0

    def getClusters(self):
        """Function returning all clusters"""
        return self.clusters
    
    def getMainCluster(self):
        """Function returning main clusters"""
        return self.main_cluster



"""Module providing finding and extraction of face from image."""

import cv2
import numpy as np
import mediapipe as mp
import eyeGestures.eye as eye


class FaceFinder:

    def __init__(self, static_image_mode=True):
        self.mp_face_mesh = mp.solutions.face_mesh.FaceMesh(
            refine_landmarks=True,
            static_image_mode=static_image_mode,
            min_detection_confidence=0.5,
            min_tracking_confidence=0.5
        )

    def find(self, image):

        assert (len(image.shape) > 2)

        try:
            face_mesh = self.mp_face_mesh.process(
                cv2.cvtColor(image, cv2.COLOR_BGR2RGB))

            return face_mesh
        except Exception as e:
            print(f"Exception in FaceFinder: {e}")
            return None


class Face:

    def __init__(self):
        self.eyeLeft = eye.Eye(0)
        self.eyeRight = eye.Eye(1)
        self.landmarks = None

    def getBoundingBox(self):
        if self.landmarks is not None:
            margin = 0
            min_x = np.min(self.landmarks[:, 0]) - margin
            max_x = np.max(self.landmarks[:, 0]) + margin
            min_y = np.min(self.landmarks[:, 1]) - margin
            max_y = np.max(self.landmarks[:, 1]) + margin

            width = int((max_x - min_x))
            height = int((max_y - min_y))
            x = int(min_x)
            y = int(min_y)
            return (x, y, width, height)
        return (0, 0, 0, 0)

    def getLeftEye(self):
        return self.eyeLeft

    def getRightEye(self):
        return self.eyeRight

    def getLandmarks(self):
        return self.landmarks

    def _landmarks(self, face):

        __complex_landmark_points = face.multi_face_landmarks
        __complex_landmarks = __complex_landmark_points[0].landmark

        __face_landmarks = []
        for landmark in __complex_landmarks:
            __face_landmarks.append((
                landmark.x * self.image_w,
                landmark.y * self.image_h))

        return np.array(__face_landmarks)

    def process(self, image, face):
        try:
            self.face = face
            self.image_h, self.image_w, _ = image.shape
            self.landmarks = self._landmarks(self.face)
            # self.nose = nose.Nose(image,self.landmarks,self.getBoundingBox())

            x, y, _, _ = self.getBoundingBox()
            offset = np.array((x, y))
            # offset = offset - self.nose.getHeadTiltOffset()

            self.eyeLeft.update(image, self.landmarks, offset)
            self.eyeRight.update(image, self.landmarks, offset)
        except Exception as e:
            return None



"""Module providing a processing pupil position to gaze direction."""

import numpy as np


class EyeProcessor:
    """Class providing processing for eye to obtain gaze direction."""

    def __init__(self, scale_w=250, scale_h=250):
        self.scale_w = scale_w
        self.scale_h = scale_h
        self.min_x = 0
        self.max_x = 0
        self.min_y = 0
        self.max_y = 0
        self.pupil = None
        self.landmarks = None

    def append(self, pupil: (int, int), landmarks: np.ndarray, pupilBuffor):
        """Function appending new pupil point to tracker."""

        self.pupil = pupil
        self.landmarks = landmarks

        # get center:
        margin = 5
        self.min_x = np.min(self.landmarks[:, 0]) - margin
        self.max_x = np.max(self.landmarks[:, 0]) + margin
        self.min_y = np.min(self.landmarks[:, 1]) - margin
        self.max_y = np.max(self.landmarks[:, 1]) + margin

        assert self.pupil[0] > self.min_x
        assert self.pupil[1] > self.min_y

        width = self.max_x - self.min_x
        height = (self.max_y - self.min_y)/2

        pupilBuffor.add(
            self.__convertPoint(self.pupil,
                                width=self.scale_w, height=self.scale_h,
                                scale_w=width, scale_h=height,
                                offset=(self.min_x, self.min_y)))

    def __convertPoint(self, point,
                       width=1.0,
                       height=1.0,
                       scale_w=1.0,
                       scale_h=1.0,
                       offset=(0.0, 0.0)):
        (min_x, min_y) = offset
        x = int(((point[0]-min_x)/scale_w)*width)
        y = int(((point[1]-min_y)/scale_h)*height)
        return (x, y)

    def getAvgPupil(self, width, height, pupilBuffor):
        """Function returning average point based on tracked."""

        if not width is None and not height is None:
            _retPupil = self.__convertPoint(pupilBuffor.getAvg(),
                                            width=width, height=height,
                                            scale_w=self.scale_w, scale_h=self.scale_h)
        else:
            _retPupil = pupilBuffor.getAvg()

        return _retPupil
